/* FILE: eulerevolve.cc                 -*-Mode: c++-*-
 *
 * Concrete evolver class, using simple forward Euler steps
 *
 */

#include <float.h>

#include "nb.h"
#include "director.h"
#include "driver.h"
#include "simstate.h"
#include "eulerevolve.h"
#include "key.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_EulerEvolve);

/* End includes */

// Constructor
Oxs_EulerEvolve::Oxs_EulerEvolve(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Evolver(name,newdtr,safe_interp,argstr),
    energy_accum_count_limit(25),
    next_timestep(0.),state_id(0),max_dm_dt(0.),dE_dt(0.),
    total_energy(0.),total_energy_change(0.),
    max_dm_dt_output(this,InstanceName(),"Max dm/dt","deg/ns",0,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr),
    dE_dt_output(this,InstanceName(),"dE/dt","J/s",0,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr),
    total_energy_output(this,InstanceName(),"Total energy","J",0,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr),
    dm_dt_output(this,InstanceName(),"dm/dt","rad/s",1,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr),
    total_H_field_output(this,InstanceName(),"H","A/m",1,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr)
{
  // Process arguments
  allowed_error = GetRealInitValue("error",90.0);
  if(allowed_error<=0.0) allowed_error=1.0;
  allowed_error *= PI*1e9/180.;  // Convert from deg/ns to rad/s

  step_aggression = GetRealInitValue("aggression",0.9);

  gamma = GetRealInitValue("gamma",2.21e5);

  do_precess = GetIntInitValue("do_precess",1);

  start_dm = GetRealInitValue("start_dm",0.01);

  alpha = GetRealInitValue("alpha"); // alpha is required input

  VerifyAllInitArgsUsed();
}

BOOL Oxs_EulerEvolve::Init()
{
  next_timestep=0.0;
  state_id=0;   // Mark as invalid state
  max_dm_dt=-1; // Dummy value
  dE_dt=0;      // Dummy value
  total_energy=0.0;
  total_energy_change=0.0; // Dummy value
  energy_accum_count=energy_accum_count_limit; // Force cold count
  /// on first pass
  return 1;
}

Oxs_EulerEvolve::~Oxs_EulerEvolve()
{}



void Oxs_EulerEvolve::Calculate_dm_dt
(const Oxs_Mesh& mesh_,
 const Oxs_MeshValue<REAL8m>& Ms_,
 const Oxs_MeshValue<ThreeVector>& mxH_,
 const Oxs_MeshValue<ThreeVector>& spin_,
 Oxs_MeshValue<ThreeVector>& dm_dt_,
 REAL8m& max_dm_dt_,REAL8m& dE_dt_) const
{ // Imports: mesh_, Ms_, mxH_, spin_
  // Exports: dm_dt_, max_dm_dt_, dE_dt_
  UINT4m size = mesh_.Size();  // Assume all imports are compatible
  REAL8m coef1 = -fabs(gamma);
  REAL8m coef2 = -fabs(alpha);
  ThreeVector scratch;
  REAL8m max_dm_dt_sq=0.0;
  REAL8m dm_dt_sq;
  REAL8m dE_dt_sum=0.0;
  dm_dt_.AdjustSize(&mesh_);
  for(UINT4m i=0;i<size;i++) {
    if(Ms_[i]==0) {
      dm_dt_[i].Set(0.0,0.0,0.0);
    } else {
      scratch = mxH_[i];
      dE_dt_sum += scratch.MagSq() * Ms_[i] * mesh_.Volume(i);
      scratch *= coef1;
      if(do_precess) {
	dm_dt_[i] = scratch;
      } else {
	dm_dt_[i].Set(0.0,0.0,0.0);
      }
      scratch ^= spin_[i]; // -|gamma|((mxH)xm)
      scratch *= coef2;  // |alpha.gamma|((mxH)xm) = -|alpha.gamma|(mx(mxH))
      dm_dt_[i] += scratch;
      dm_dt_sq = dm_dt_[i].MagSq();
      if(dm_dt_sq>max_dm_dt_sq) max_dm_dt_sq=dm_dt_sq;
    }
  }
  max_dm_dt_ = sqrt(max_dm_dt_sq);
  dE_dt_ = -1 * MU0 * fabs(gamma*alpha) * dE_dt_sum;
  return;
}




BOOL
Oxs_EulerEvolve::Step(const Oxs_Driver& driver,
		      Oxs_ConstKey<Oxs_SimState> current_state,
		      Oxs_Key<Oxs_SimState>& next_state)
{
  const REAL8m max_step_increase = 1.25;
  const REAL8m max_step_decrease = 0.5;

  UINT4m size,i; // Mesh size and indexing variable

  const Oxs_SimState& cstate = current_state.GetReadReference();
  Oxs_SimState& workstate = next_state.GetWriteReference();

  if(cstate.mesh->Id() != workstate.mesh->Id()) {
    throw Oxs_Ext::Error
      ("EulerEvolve::Step: Oxs_Mesh not fixed across steps.");
  }

  REAL8m stepsize = next_timestep;
  if( cstate.Id() != state_id || stepsize<=0.0) {
    // Cached data out of data
    director->GetEnergyDensity(cstate,energy,&mxH,NULL);
    Calculate_dm_dt(*(cstate.mesh),*(cstate.Ms),
		    mxH,cstate.spin,
		    dm_dt,max_dm_dt,dE_dt);
    if(start_dm < sqrt(DBL_MAX/4) * max_dm_dt) {
      stepsize = start_dm / max_dm_dt;
    } else {
      stepsize = sqrt(DBL_MAX/4);
    }
    state_id = cstate.Id();
  }

   // Negotiate with driver over size of next step
  workstate.last_timestep=stepsize;
  driver.FillStatePartB(workstate);
  BOOL forcestep=0;
  if(workstate.last_timestep>stepsize) {
    // driver wants to force this stepsize
    forcestep=1;
  }
  stepsize = workstate.last_timestep;

  if(stepsize*max_dm_dt<REAL8_EPSILON) {
    // Step too small to have effect, so force acceptance.
    forcestep=1;
  }

  workstate.elapsed_time = cstate.elapsed_time + stepsize;
  workstate.iteration_count = cstate.iteration_count + 1;
  workstate.stage_iteration_count = cstate.stage_iteration_count + 1;

  // Put new spin configuration in next_state
  workstate.spin.AdjustSize(workstate.mesh); // Safety
  size = workstate.spin.Size();
  ThreeVector tempspin;
  for(i=0;i<size;++i) {
    tempspin = dm_dt[i];
    tempspin *= stepsize;

    // For improved accuracy, adjust step vector so that
    // to first order m0 + adjusted_step = v/|v| where
    // v = m0 + step.
    REAL8m adj = 0.5 * tempspin.MagSq();
    tempspin -= adj*cstate.spin[i];
    tempspin *= 1.0/(1.0+adj);

    tempspin += cstate.spin[i];
    tempspin.SetMag(1.0);
    workstate.spin[i] = tempspin;
  }
  const Oxs_SimState& nstate
    = next_state.GetReadReference();  // Release write lock

  //  Calculate delta E
  Oxs_MeshValue<ThreeVector>* H_ptr = &new_H;
  if(total_H_field_output.GetCacheRequestCount()>0) {
    H_ptr = &total_H_field_output.cache.value;
    total_H_field_output.cache.state_id = 0;
  }
  director->GetEnergyDensity(nstate,new_energy,&mxH,H_ptr);
  if(total_H_field_output.GetCacheRequestCount()>0) {
    total_H_field_output.cache.state_id = nstate.Id();
  }

  REAL8m dE=0.0;
  for(i=0;i<size;++i) {
    dE += (new_energy[i] - energy[i]) * nstate.mesh->Volume(i);
  }

  // Get error estimate.  See step size adjustment discussion in
  // MJD Notes II, p72 (18-Jan-2001).
  REAL8m new_max_dm_dt,new_dE_dt;
  Calculate_dm_dt(*(nstate.mesh),
		  *(nstate.Ms),
		  mxH,nstate.spin,
		  new_dm_dt,new_max_dm_dt,new_dE_dt);
  REAL8m max_error=0;
  for(i=0;i<size;++i) {
    ThreeVector temp = dm_dt[i];
    temp -= new_dm_dt[i];
    REAL8m temp_error = temp.MagSq();
    if(temp_error>max_error) max_error = temp_error;
  }
  max_error = sqrt(max_error)/2.0; // Actual (local) error
  /// estimate is max_error * stepsize


  // Estimate ideal stepsize using quadratic fit to results.
  // First calculate dE/dM|t1 * dm/dt|t0 (see NOTES II, p66,
  // 13-Dec-2000):
  REAL8m dE_dt_1 = 0;
  for(i=0;i<size;++i) {
    dE_dt_1 += ((*H_ptr)[i] * dm_dt[i]) * nstate.mesh->Volume(i);
  }
  dE_dt_1 *= -1; // NOTES say this should be *MU0?

#ifdef OLDE_CODE
  // Now, use formula (5) in NOTES II, p69, 13-Dec-2000, to
  // estimate ideal step size.
  REAL8m a = 0.5*(dE_dt_1 - dE_dt);
  REAL8m b = 0.2*dE*stepsize + 0.9*dE_dt - 0.1*dE_dt_1;
  REAL8m t_ideal = 1.0; // This is relative to current step
  if(a>0.0) {
    t_ideal = 0.5;
    if(-b>a) {
      if(-b<2*a*max_step_increase) t_ideal = -b/(2*a);
      else                         t_ideal = max_step_increase;
    }
  }
#else
  // This seems to work slightly better: Fit using E at
  // the two ends, and just dE_dt_1 (i.e., ignore dE_dt).
  REAL8m a = dE_dt_1*stepsize - dE;
  REAL8m b = 2*dE - dE_dt_1*stepsize;
  REAL8m t_ideal = 1.0; // This is relative to current step
  if(a>0.0) {
    t_ideal = max_step_decrease;
    if(-b>2*max_step_decrease*a) {
      if(-b<2*max_step_increase*a) t_ideal = -b/(2*a);
      else                         t_ideal = max_step_increase;
    }
  }
#endif

  // Check step and adjust next_timestep
  if(!forcestep &&
     (dE>REAL8_EPSILON*fabs(total_energy)
      || max_error>allowed_error || t_ideal<0.75)) {
    if(dE>0.0) next_timestep=stepsize*max_step_decrease;
    else       {
      next_timestep = 0.9*allowed_error/max_error;
      if(next_timestep>t_ideal) next_timestep = t_ideal;
      if(next_timestep<max_step_decrease)
	next_timestep=max_step_decrease;
      next_timestep *= stepsize;
    }
    return 0;  // Reject step
  }

  // Otherwise, accept step.  Calculate next step size using
  // two methods.  First, estimate what size step would just
  // meet the error restriction (with "aggression" safety
  // margin).  Then compare that size to the "ideal" size
  // calculated using the quadratic fit to E(t), and use the
  // smaller.
  state_id=0; // Safety
  if(step_aggression*allowed_error<max_error*max_step_increase)
    next_timestep = step_aggression*allowed_error/max_error;
  else
    next_timestep = max_step_increase;
  if(next_timestep>t_ideal) next_timestep = t_ideal;
  if(next_timestep<max_step_decrease)
    next_timestep=max_step_decrease; // Safety
  next_timestep *= stepsize;
  dm_dt.Swap(new_dm_dt);
  max_dm_dt=new_max_dm_dt;
  dE_dt=new_dE_dt;
  if(energy_accum_count<energy_accum_count_limit) {
    total_energy += dE;
    ++energy_accum_count;
  } else {
    total_energy=0.0;
    for(i=0;i<size;++i) {
      total_energy += new_energy[i] * nstate.mesh->Volume(i);
    }
    energy_accum_count=0;
  }
  total_energy_change = dE;
  energy.Swap(new_energy);
  state_id = nstate.Id();

  return 1;  // Good step
}

void Oxs_EulerEvolve::UpdateDerivedOutputs(const Oxs_SimState& state)
{ // This routine fills all the Oxs_EulerEvolve Oxs_ScalarOutput's to
  // the appropriate value based on the import "state".  Also any of
  // Oxs_VectorOutput's that have CacheRequest enabled are filled. 
  max_dm_dt_output.cache.state_id
    = dE_dt_output.cache.state_id
    = total_energy_output.cache.state_id = 0;
  if(state.Id()==state_id) {
    // Use previously calculated values
    max_dm_dt_output.cache.value=max_dm_dt*(180e-9/PI);
             /// Convert from radians/second to deg/ns
    dE_dt_output.cache.value=dE_dt;
    total_energy_output.cache.value=total_energy;
    if(dm_dt_output.GetCacheRequestCount()>0) {
      dm_dt_output.cache.value=dm_dt;
      dm_dt_output.cache.state_id=state_id;
    }
  } else {
    // Calculate from scratch
    UINT4m size=state.mesh->Size();
    Oxs_MeshValue<ThreeVector>* H_ptr = NULL;
    if(total_H_field_output.GetCacheRequestCount()>0
       && total_H_field_output.cache.state_id != state.Id()) {
      H_ptr = &total_H_field_output.cache.value;
      total_H_field_output.cache.state_id = 0;
    }
    director->GetEnergyDensity(state,new_energy,&mxH,H_ptr);
    if(total_H_field_output.GetCacheRequestCount()>0
       && total_H_field_output.cache.state_id != state.Id()) {
      total_H_field_output.cache.state_id = state.Id();
    }
    REAL8m energy_sum=0.;
    for(UINT4m i=0;i<size;++i) {
      energy_sum += new_energy[i] * state.mesh->Volume(i);
    }
    total_energy_output.cache.value=energy_sum;

    Oxs_MeshValue<ThreeVector>* dm_dt_ptr = &dm_dt;
    if(dm_dt_output.GetCacheRequestCount()>0
       && dm_dt_output.cache.state_id != state.Id()) {
      dm_dt_ptr = &dm_dt_output.cache.value;
      dm_dt_output.cache.state_id=0;
    }
    Calculate_dm_dt(*(state.mesh),*(state.Ms),mxH,state.spin,
		    *dm_dt_ptr,max_dm_dt_output.cache.value,
		    dE_dt_output.cache.value);
    if(dm_dt_output.GetCacheRequestCount()>0
       && dm_dt_output.cache.state_id != state.Id()) {
      dm_dt_output.cache.state_id=state.Id();
    }
    max_dm_dt_output.cache.value *= 180e-9/PI;
      /// Convert from radians/second to deg/ns
  }
  max_dm_dt_output.cache.state_id
    = dE_dt_output.cache.state_id
    = total_energy_output.cache.state_id = state.Id();

  if(total_H_field_output.GetCacheRequestCount()>0 
     && total_H_field_output.cache.state_id != state.Id()) {
    director->GetEnergyDensity(state,new_energy,NULL,
			       &total_H_field_output.cache.value);
    total_H_field_output.cache.state_id = state.Id();
  }
}


