/* FILE: standarddriver.cc            -*-Mode: c++-*-
 *
 * Example concrete Oxs_Driver class.
 *
 */

#include "nb.h"
#include "standarddriver.h"
#include "director.h"
#include "simstate.h"
#include "evolver.h"
#include "key.h"
#include "mesh.h"
#include "scalarfieldinit.h"
#include "vectorfieldinit.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_StandardDriver);

/* End includes */

#define BUF_SIZE 1024
static char buf[BUF_SIZE];	// Buffer for building interpreter results

// Constructor
#define OSO_INIT(name,descript,units) \
   name##_output(this,InstanceName(),descript,units,0, \
           &Oxs_StandardDriver::Fill__##name##_output,newdtr)
Oxs_StandardDriver::Oxs_StandardDriver(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp_, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Driver(name,newdtr,safe_interp_,argstr),
    safe_interp(safe_interp_),
    max_dm_dt_obj_ptr(NULL),
    OSO_INIT(iteration_count,"Iteration",""),
    OSO_INIT(stage_iteration_count,"Stage iteration",""),
    OSO_INIT(stage_number,"Stage",""),
    OSO_INIT(last_timestep,"Last time step","s"),
    OSO_INIT(simulation_time,"Simulation time","s"),
    magnetization_output(this,InstanceName(),"Magnetization","A/m",1,
			 &Oxs_StandardDriver::Fill__magnetization_output,
			 newdtr),
    aveMx_output(this,InstanceName(),"Mx","A/m",1,
			 &Oxs_StandardDriver::Fill__aveM_output,
			 newdtr),
    aveMy_output(this,InstanceName(),"My","A/m",1,
			 &Oxs_StandardDriver::Fill__aveM_output,
			 newdtr),
    aveMz_output(this,InstanceName(),"Mz","A/m",1,
			 &Oxs_StandardDriver::Fill__aveM_output,
			 newdtr)
{
  // Process arguments
  max_timestep=GetRealInitValue("max_timestep");
  if(max_timestep<=0.0) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid MIF input block detected for object %s:"
		" Specified max time step is %g",InstanceName(),
		max_timestep);
    throw Oxs_Ext::Error(buf);
  }

  min_timestep=GetRealInitValue("min_timestep");

  stopping_dm_dt=GetRealInitValue("stopping_dm_dt");

  const vector<string>*  vsptr; // Temp pointer
  number_of_stages=1;  // Default value: 1 stage
  vsptr = FindInitValue("number_of_stages");
  if(vsptr!=NULL) {
    number_of_stages=GetIntInitValue("number_of_stages");
  }

  stage_iteration_limit=0;  // Default value: no limit
  vsptr = FindInitValue("stage_iteration_limit");
  if(vsptr!=NULL) {
    stage_iteration_limit=GetIntInitValue("stage_iteration_limit");
  }

  total_iteration_limit=0;  // Default value: no limit
  vsptr = FindInitValue("total_iteration_limit");
  if(vsptr!=NULL) {
    total_iteration_limit=GetIntInitValue("total_iteration_limit");
  }
 
  CheckInitValueParamCount("mesh",1);
  Oxs_Mesh* mesh = dynamic_cast<Oxs_Mesh*>
    (director->FindExtObject((*FindInitValue("mesh"))[0]));
  if(mesh==NULL) {
    string msg="Unable to identify Oxs_Mesh object with name "
      + (*FindInitValue("mesh"))[0];
    throw Oxs_Ext::Error(msg.c_str());
  }
  DeleteInitValue("mesh");
  meshkey.Set(mesh);  // Sets a dep lock

  auto_ptr<Oxs_ScalarFieldInit>
    Msinit(NewScalarFieldInit(safe_interp,"Ms"));

  auto_ptr<Oxs_VectorFieldInit> foo(NewVectorFieldInit(safe_interp,"m0"));
  m0 = foo;
  /// Stroustrup specifies an auto_ptr .reset() member function,
  /// but that is apparently not supported by some common STL's.
  /// The above creation of a temporary (foo) is a workaround.

  VerifyAllInitArgsUsed();

  // Fill Ms and Ms_inverse array
  Msinit->FillMeshValue(mesh,Ms);
  Ms_inverse.AdjustSize(mesh);
  for(UINT4m i=0;i<mesh->Size();i++) {
    if(Ms[i]!=0.0) Ms_inverse[i]=1.0/Ms[i];
    else           Ms_inverse[i]=0.0;
  }

  // Reserve space for initial state (see GetInitialState() below)
  director->ReserveSimulationStateRequest(1);
}

Oxs_ConstKey<Oxs_SimState>
Oxs_StandardDriver::GetInitialState() const
{
  Oxs_Key<Oxs_SimState> initial_state;
  director->GetNewSimulationState(initial_state);
  Oxs_SimState& istate = initial_state.GetWriteReference();
  istate.iteration_count       =  0;
  istate.stage_iteration_count =  0;
  istate.stage_number          =  0;
  istate.last_timestep         =  0.0;
  istate.elapsed_time          =  0.0;

  istate.mesh = meshkey.GetPtr();

  istate.Ms = &Ms;
  istate.Ms_inverse = &Ms_inverse;
  m0->FillMeshValue(istate.mesh,istate.spin);

  // Insure that all spins are unit vectors
  UINT4m size = istate.spin.Size();
  for(UINT4m i=0;i<size;i++) istate.spin[i].SetMag(1.0);

  initial_state.GetReadReference();  // Release write lock.
  /// The read lock will be automatically released when the
  /// key "initial_state" is destroyed.

  return initial_state;
}


BOOL Oxs_StandardDriver::Init()
{ 
  Init_Driver();  // Run init routine in parent.  This will call
  /// Oxs_StandardDriver::GetInitialState().

  // Get pointer to output object providing max dm/dt data
  max_dm_dt_obj_ptr =  director->FindOutputObject("Max dm/dt");
  if(max_dm_dt_obj_ptr==NULL) {
    string msg=string("Error in ") + string(InstanceName())
      + string(": Unable to identify unique Max dm/dt output object");
    throw Oxs_Ext::Error(msg.c_str());
  }
  return 1;
}

Oxs_StandardDriver::~Oxs_StandardDriver()
{}

BOOL
Oxs_StandardDriver::StageDone(const Oxs_SimState& state, string& result) const
{
  REAL8m max_dm_dt=0.;
  if( stage_iteration_limit>0
      && state.stage_iteration_count >= stage_iteration_limit) goto done;
  if( total_iteration_limit>0
      && state.iteration_count >= total_iteration_limit ) goto done;
  if(max_dm_dt_obj_ptr->Output(&state,safe_interp,0,NULL) != TCL_OK) {
    string msg=string("Error in ") + string(InstanceName())
      + string(": Unable to obtain Max dm/dt output: ");
    msg += string(Tcl_GetStringResult(safe_interp));
    throw Oxs_Ext::Error(msg.c_str());
  }
  max_dm_dt = Nb_Atof(Tcl_GetStringResult(safe_interp));
  if(max_dm_dt <= stopping_dm_dt) goto done;
  return 0;
done:
  result.append("Oc_EventHandler Generate Oxs Stage\n");
  return 1;
}

BOOL Oxs_StandardDriver::Done(const Oxs_SimState& state, string& result) const
{
  if( total_iteration_limit>0
      && state.iteration_count >= total_iteration_limit ) goto done;
  if( number_of_stages >0
      && state.stage_number >= number_of_stages ) goto done;
  return 0;
done:
  result.append("Oc_EventHandler Generate Oxs Done\n");
  return 1;
}

void
Oxs_StandardDriver::NewState(const Oxs_SimState& state, string& result) const
{
  Oc_Snprintf(buf, sizeof(buf), 
              "Oc_EventHandler Generate Oxs Step -step %d -stage %d\n",
              state.iteration_count, state.stage_number);
  result.append(buf);
}

void Oxs_StandardDriver::FillStatePartA
(const Oxs_SimState& old_state,
 Oxs_SimState& new_state) const
{
  new_state.stage_number = old_state.stage_number;
  new_state.Ms           = old_state.Ms;
  new_state.Ms_inverse   = old_state.Ms_inverse;
  new_state.mesh         = old_state.mesh;
  new_state.spin.AdjustSize(new_state.mesh);
}

void Oxs_StandardDriver::FillStatePartB(Oxs_SimState& work_state) const
{
  if(work_state.last_timestep<min_timestep)
    work_state.last_timestep=min_timestep;
  if(work_state.last_timestep>max_timestep)
    work_state.last_timestep=max_timestep;
}

////////////////////////////////////////////////////////////////////////
// State-based outputs, maintained by the driver.  These are
// conceptually public, but are specified private to force
// clients to use the output_map interface in Oxs_Director.

#define OSO_FUNC(NAME) \
void Oxs_StandardDriver::Fill__##NAME##_output(const Oxs_SimState& state) \
{ NAME##_output.cache.state_id=state.Id(); \
  NAME##_output.cache.value=state.NAME; }

OSO_FUNC(iteration_count)
OSO_FUNC(stage_iteration_count)
OSO_FUNC(stage_number)
OSO_FUNC(last_timestep)
void
Oxs_StandardDriver::Fill__simulation_time_output(const Oxs_SimState& state)
{
  simulation_time_output.cache.state_id = state.Id();
  simulation_time_output.cache.value = state.elapsed_time;
}

void
Oxs_StandardDriver::Fill__magnetization_output(const Oxs_SimState& state)
{
  magnetization_output.cache.state_id=0;
  magnetization_output.cache.value.AdjustSize(state.mesh);
  UINT4m size=state.mesh->Size();
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  Oxs_MeshValue<ThreeVector>& mag = magnetization_output.cache.value;
  for(UINT4m i=0;i<size;i++) {
    mag[i] = spin[i];
    mag[i] *= Ms[i];
  }
  magnetization_output.cache.state_id=state.Id();
}

void
Oxs_StandardDriver::Fill__aveM_output(const Oxs_SimState& state)
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  UINT4m size=state.mesh->Size();

  if(aveMx_output.GetCacheRequestCount()>0 &&
     aveMy_output.GetCacheRequestCount()>0 &&
     aveMz_output.GetCacheRequestCount()>0) {
    // Usual case: All three components desired
    aveMx_output.cache.state_id=0;
    aveMy_output.cache.state_id=0;
    aveMz_output.cache.state_id=0;
    REAL8m Mx=0.0;
    REAL8m My=0.0;
    REAL8m Mz=0.0;
    if(size>0) {
      UINT4m i=0;
      do {
	REAL8m sat_mag = Ms[i];
	Mx += sat_mag*(spin[i].x);
	My += sat_mag*(spin[i].y);
	Mz += sat_mag*(spin[i].z);
	++i;
      } while(i<size);
      Mx/=size;
      My/=size;
      Mz/=size;
    }
    aveMx_output.cache.value=Mx;
    aveMx_output.cache.state_id=state.Id();
    aveMy_output.cache.value=My;
    aveMy_output.cache.state_id=state.Id();
    aveMz_output.cache.value=Mz;
    aveMz_output.cache.state_id=state.Id();
  } else {
    // Calculate components on a case-by-case basis
    if(aveMx_output.GetCacheRequestCount()>0) {
      aveMx_output.cache.state_id=0;
      REAL8m Mx=0.0;
      if(size>0) {
	UINT4m i=0;
	do {
	  Mx += Ms[i]*(spin[i].x);
	  ++i;
	} while(i<size);
	Mx/=size;
      }
      aveMx_output.cache.value=Mx;
      aveMx_output.cache.state_id=state.Id();
    }

    if(aveMy_output.GetCacheRequestCount()>0) {
      aveMy_output.cache.state_id=0;
      REAL8m My=0.0;
      if(size>0) {
	UINT4m i=0;
	do {
	  My += Ms[i]*(spin[i].y);
	  ++i;
	} while(i<size);
	My/=size;
      }
      aveMy_output.cache.value=My;
      aveMy_output.cache.state_id=state.Id();
    }

    if(aveMz_output.GetCacheRequestCount()>0) {
      aveMz_output.cache.state_id=0;
      REAL8m Mz=0.0;
      if(size>0) {
	UINT4m i=0;
	do {
	  Mz += Ms[i]*(spin[i].z);
	  ++i;
	} while(i<size);
	Mz/=size;
      }
      aveMz_output.cache.value=Mz;
      aveMz_output.cache.state_id=state.Id();
    }
  }
}
